!
! Module:dss_translate
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This is a temporary module which returns the old DSS calls and data types but in projected new names.
! 
! $Id: dss_translate.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module dss_translate

  use dal
  use errorHandling
  use dss

  implicit none 

  private 

  public readDss
  public writeDss
  public releaseDss
  public deleteDss
  public dssHasAxis
  private dissectCommaDelimitedStr
  public dssToSelectExpr

  public numDssComponents
  public getDssComponent

  public numDssGroups

!**** the following should be seen as temporary since they all assume that all elements have only 1 filter.
  public getDssFilter
  public getDssFilterType
  public getDssFilterAxisName
  public getDssFilterAxisUnit
  public getDssFilterAxisNames
  public getDssFilterBlockName
  public getDssFilterBitMasks
  public getDssFilterRanges




  type, public :: DssComponentT
!    character(20)         :: name
!    logical(bool)         :: ignore
!    type(RangeT), pointer :: ranges(:)
  end type DssComponentT

  type, public :: DssFilterT
!    type(RangeT), pointer :: ranges(:)
!    integer, pointer :: compNums(:)
  end type DssFilterT

contains 

  !---------------------------------------------------------------------
  function readDss(inBlock) result(outDss)
  !---------------------------------------------------------------------

    implicit none

    type(BlockT), intent(in) :: inBlock
    type(DssT)               :: outDss

    !-------------------------------------------------------------------
    ! Start

    outDss = dataSubspace(inBlock)

  end function readDss

  !---------------------------------------------------------------------
  subroutine writeDss(inDss, outBlock)
  !---------------------------------------------------------------------

    implicit none

    type(DssT),   intent(in) :: inDss
    type(BlockT), intent(in) :: outBlock

    !-------------------------------------------------------------------
    ! Start

    call dssWrite(inDss, outBlock)

  end subroutine writeDss

  !---------------------------------------------------------------------
  subroutine releaseDss(inDss)
  !---------------------------------------------------------------------

    implicit none

    type(DssT), intent(in) :: inDss

    !-------------------------------------------------------------------
    ! Start

    call dssRelease(inDss)

  end subroutine releaseDss

  !---------------------------------------------------------------------
  subroutine deleteDss(inBlock)
  !---------------------------------------------------------------------

    implicit none

    type(BlockT), intent(in) :: inBlock

    !-------------------------------------------------------------------
    ! Start

    call dssDelete(inBlock)

  end subroutine deleteDss

  !---------------------------------------------------------------------
  logical function dssHasAxis_0(inDss, axisName)
  !---------------------------------------------------------------------

    implicit none

    type(DssT),   intent(in) :: inDss
    character(*), intent(in) :: axisName

    integer :: numComps, compNum, filterType
    type(DScompT) :: dssComp

    !-------------------------------------------------------------------
    ! Start

    dssHasAxis = .false. ! default
    numComps = dssComponents(inDss)
  if (numComps <= 0) return

    do compNum = 0, numComps - 1
      dssComp = dssComponent(compNum)
      do filterType = RANGEFILTERT, REGIONBLOCKFILTERT
        if (dssHasFilter(dssComp, axisName, filterType)) then
          dssHasAxis = .true.
  return
        end if
      end do
    end do

  end function dssHasAxis_0

  !---------------------------------------------------------------------
  logical function dssHasAxis(inDss, axisName)
  !---------------------------------------------------------------------

    implicit none

    type(DssT),   intent(in) :: inDss
    character(*), intent(in) :: axisName

    integer :: numComps, compNum, numFilts, filtNum, filterType, i
    type(DScompT) :: dssComp
    type(DSfilterT) :: dssFilt
    character(80) :: compStr
    character(20), pointer :: strList(:)

    !-------------------------------------------------------------------
    ! Start

    dssHasAxis = .false. ! default
    numComps = dssComponents(inDss)
  if (numComps <= 0) return

    do compNum = 0, numComps - 1
      dssComp = dssComponent(compNum)
      numFilts = dssFilters(dssComp)
    if (numFilts <= 0) cycle

      if (dssHasFilter(dssComp, axisName, RANGEFILTERT)) then
        dssHasAxis = .true.
  return
      end if

      if (dssHasFilter(dssComp, axisName, BITMASKFILTERT)) then
        dssHasAxis = .true.
  return
      end if

      if (dssHasFilter(dssComp, axisName, RANGEBLOCKFILTERT)) then
        dssHasAxis = .true.
  return
      end if

      if (dssHasFilter(dssComp, axisName, GTIFILTERT)) then
        dssHasAxis = .true.
  return
      end if

      do filtNum = 0, numFilts - 1
        dssFilt = dssFilter(dssComp, filtNum)
        filterType = dssFilterType(dssFilt)

        if ( filterType == REGIONFILTERT&
        .or. filterType == MASKFILTERT&
        .or. filterType == REGIONBLOCKFILTERT)&
        then
          compStr = dssFilterComps(dssFilt)
          strList => dissectCommaDelimitedStr(compStr)
          do i = 1, size(strList)
            if (strList(i) == axisName) then
              dssHasAxis = .true.
  return
            end if
          end do
          deallocate(strList)
        end if
      end do
    end do

  end function dssHasAxis

  !---------------------------------------------------------------------
  function dissectCommaDelimitedStr(inStr) result(strList)
  !---------------------------------------------------------------------
  ! Note: the calling program should deallocate the pointer strList after use.

!*** See if can have size of elements = *

    implicit none

    character(*),  intent(in) :: inStr
    character(20), pointer    :: strList(:)

    integer :: startI, numItems, nextCommaI, itemNum

    !-------------------------------------------------------------------
    ! Start

    ! First, find out how many items there will be:

    startI = 1
    numItems = 1
    do
      nextCommaI = index(inStr(startI:), ',')
    if (nextCommaI < 1) exit
      startI = startI + nextCommaI
      numItems = numItems + 1
    end do

    ! Now load up the elements:

    allocate(strList(numItems))
    startI = 1
    itemNum = 1
    do
      nextCommaI = index(inStr(startI:), ',')
    if (nextCommaI < 1) exit
      strList(itemNum) = inStr(startI:startI+nextCommaI-2)
      startI = startI + nextCommaI
      itemNum = itemNum + 1
    end do
    strList(itemNum) = inStr(startI:)

  end function dissectCommaDelimitedStr

  !---------------------------------------------------------------------
  function dssToSelectExpr(inDss) result(selectionExpr)
  !---------------------------------------------------------------------

    implicit none

    type(DssT), intent(in) :: inDss
    character(1024)        :: selectionExpr

    !-------------------------------------------------------------------
    ! Start

    selectionExpr = ''
    selectionExpr = selectExpression(inDss)

  end function dssToSelectExpr


  !---------------------------------------------------------------------
  integer function numDssComponents(inDss)
  !---------------------------------------------------------------------

    implicit none

    type(DssT), intent(in) :: inDss

    !-------------------------------------------------------------------
    ! Start

    numDssComponents = dssComponents(inDss)

  end function numDssComponents

  !---------------------------------------------------------------------
  function getDssComponent(inDss, compNum) result(dssComp)
  !---------------------------------------------------------------------

    implicit none

    type(DssT), intent(in) :: inDss
    integer,    intent(in) :: compNum
    type(DssComponentT)    :: dssComp

    !-------------------------------------------------------------------
    ! Start

    numDssComps = dssComponents(inDss)
if (compNum < 1 .or. compNum > numDssComps) call error('',"")
    dssCompOld = dssComponent(inDss, compNum-1)

    numGroups = numDssGroups(inDss)

  end function getDssComponent

  !---------------------------------------------------------------------
  integer function numDssGroups
  !---------------------------------------------------------------------

    implicit none

    type(DssT), intent(in) :: inDss

    !-------------------------------------------------------------------
    ! Start

  end function numDssGroups

  !---------------------------------------------------------------------
  function getDssFilter(dssComp, groupNum) result(dssFilt)
  !---------------------------------------------------------------------

    implicit none

    type(DssComponentT), intent(in) :: dssComp
    integer,             intent(in) :: groupNum
    type(DssFilterT)                :: dssFilt

    !-------------------------------------------------------------------
    ! Start

  end function getDssFilter

  !---------------------------------------------------------------------
  integer function getDssFilterType(dssFilt)
  !---------------------------------------------------------------------

    implicit none

    type(DssFilterT), intent(in) :: dssFilt

    !-------------------------------------------------------------------
    ! Start

  end function getDssFilterType

  !---------------------------------------------------------------------
  function getDssFilterAxisName(dssFilt) result(axisName)
  !---------------------------------------------------------------------

    implicit none

    type(DssFilterT), intent(in) :: dssFilt
    character(20)                :: axisName

    !-------------------------------------------------------------------
    ! Start

  end function getDssFilterAxisName

  !---------------------------------------------------------------------
  function getDssFilterAxisUnit(dssFilt) result(axisUnit)
  !---------------------------------------------------------------------

    implicit none

    type(DssFilterT), intent(in) :: dssFilt
    character(20)                :: axisUnit

    !-------------------------------------------------------------------
    ! Start

  end function getDssFilterAxisUnit

  !---------------------------------------------------------------------
  function getDssFilterAxisNames(dssFilt) result(axisNames)
  !---------------------------------------------------------------------

    implicit none

    type(DssFilterT), intent(in) :: dssFilt
    character(20),    pointer    :: axisNames(:)

    !-------------------------------------------------------------------
    ! Start

  end function getDssFilterAxisNames

  !---------------------------------------------------------------------
  function getDssFilterBlockName(dssFilt) result(blockName)
  !---------------------------------------------------------------------

    implicit none

    type(DssFilterT), intent(in) :: dssFilt
    character(80)                :: blockName

    !-------------------------------------------------------------------
    ! Start

  end function getDssFilterBlockName

  !---------------------------------------------------------------------
  function getDssFilterBitMasks(dssFilt) result(bitMasks)
  !---------------------------------------------------------------------

    implicit none

    type(DssFilterT), intent(in) :: dssFilt
    type(BitMaskT),   pointer    :: bitMasks(:)

    !-------------------------------------------------------------------
    ! Start

  end function getDssFilterBitMasks

  !---------------------------------------------------------------------
  function getDssFilterRanges(dssFilt) result(ranges)
  !---------------------------------------------------------------------

    implicit none

    type(DssFilterT), intent(in) :: dssFilt
    type(RangeT),     pointer    :: ranges(:)

    !-------------------------------------------------------------------
    ! Start

  end function getDssFilterRanges


  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module dss_translate
