!+
! Module: sort_mod
!
! Authors:  Clive Page - SSC/LUX
!          
! Description: Non-recursive stack versions of quicksort from N.Wirth's Pascal
!  book, "Algorithms + data structures = programs".
!
! Converted to Fortran90 by Alan Miller, CSIRO.
!
! Module constructed by Clive Page, Leicester University.
!
! Some changes to format, + quickSortStatic added, by I Stewart 13 May 2003.
!
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module sort_mod

  use types		! For single, double, int32 etc type definitions.


!***** use ssc_misc for swap().


  implicit none

  private 

  public quickSort
  public quickSortStatic

  interface quickSort
     module procedure quickSortStr,     quickSortStrWithIndex
     module procedure quickSortDouble,  quickSortDoubleWithIndex
     module procedure quickSortInt16
     module procedure quickSortInt32,   quickSortInt32WithIndex
     module procedure quickSortSingle,  quickSortSingleWithIndex
  end interface quickSort

  interface quickSortStatic
     module procedure quickSortStaticDouble
  end interface quickSortStatic

contains

  !---------------------------------------------------------------------
  subroutine quickSortStaticDouble(values, index)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: values(:)
    integer,      intent(out) :: index(size(values))

    real(double), allocatable :: tempValues(:)

    !-------------------------------------------------------------------
    ! Start

    allocate(tempValues(size(values)))

    tempValues = values
    call quickSort(tempValues, index)

  end subroutine quickSortStaticDouble

  !---------------------------------------------------------------------
  subroutine quickSortStr(array)
  !---------------------------------------------------------------------

    character(*), intent(INOUT)    :: array(:)
    character(LEN=len(array)) :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(30), stackr(30)

    !-------------------------------------------------------------------
    ! Start

    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit  
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                array(i) = array(j)
                array(j) = temp
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
!    return

  end subroutine quickSortStr

  !---------------------------------------------------------------------
  subroutine quickSortDouble(array)
  !---------------------------------------------------------------------

    real(double), intent(INOUT)    :: array(:)
    real(double) :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(30), stackr(30)

    !-------------------------------------------------------------------
    ! Start

    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit  
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                array(i) = array(j)
                array(j) = temp
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
!    return

  end subroutine quickSortDouble

  !---------------------------------------------------------------------
  subroutine quickSortInt16(int16Array)
  !---------------------------------------------------------------------

    integer(int16), intent(INOUT) :: int16Array(:)

    integer(int32) :: int32Array(size(int16Array))

    !-------------------------------------------------------------------
    ! Start

    int32Array = int16Array
    call quickSortInt32(int32Array)
    int16Array = int32Array

  end subroutine quickSortInt16

  !---------------------------------------------------------------------
  subroutine quickSortInt32(array)
  !---------------------------------------------------------------------

    integer(int32), intent(INOUT)    :: array(:)
    integer(int32) :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(30), stackr(30)

    !-------------------------------------------------------------------
    ! Start

    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit  
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                array(i) = array(j)
                array(j) = temp
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
!    return

  end subroutine quickSortInt32

  !---------------------------------------------------------------------
  subroutine quickSortSingle(array)
  !---------------------------------------------------------------------

    real(single), intent(INOUT)    :: array(:)
    real(single)     :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(30), stackr(30)

    !-------------------------------------------------------------------
    ! Start

    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit  
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                array(i) = array(j)
                array(j) = temp
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
!    return

  end subroutine quickSortSingle

  !---------------------------------------------------------------------
  subroutine quickSortStrWithIndex(array, order)
  !---------------------------------------------------------------------

    character(*), intent(INOUT)    :: array(:)
    integer, intent(OUT) :: order(:)  ! returns original position
    character(LEN=len(array)) :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(15), stackr(15), ww

    !-------------------------------------------------------------------
    ! Start

    do j = lbound(array,1),ubound(array,1)
       order(j) = j
    end do
    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                ww   = order(i)
                array(i) = array(j)
                order(i) = order(j)
                array(j) = temp
                order(j) = ww
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
!    return

  end subroutine quickSortStrWithIndex

  !---------------------------------------------------------------------
  subroutine quickSortDoubleWithIndex_new(array, order)
  !---------------------------------------------------------------------

    real(double), intent(INOUT)    :: array(:)
    integer, intent(OUT) :: order(:)  ! returns original position
    real(double) :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(15), stackr(15), ww

    !-------------------------------------------------------------------
    ! Start

print *, 'lbound(array,1),ubound(array,1),lbound(order,1),ubound(order,1)'&
,lbound(array,1),ubound(array,1),lbound(order,1),ubound(order,1) 

    do j = lbound(array,1),ubound(array,1)
       order(j) = j
    end do
    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

print *, 's=', s
    do
print *, ''
print *, '  start loop A'
       l = stackl(s)
       r = stackr(s)
print *, '  l = stackl(s=',s,')=', l
print *, '  r = stackr(s=',s,')=', r
       s = s - 1
print *, '  s=', s

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
print *, ''
print *, '    start loop B'
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)
print *, '    i = l=', i
print *, '    j = r=', j
print *, '    k = (l+r) / 2=', k
print *, '    x = array(k=',k,')=', x
!!
          !     REPEAT UNTIL I > J.

          do
print *, ''
print *, '      start loop C'
             do                          ! Search from lower end
print *, ''
print *, '        start loop D1 - low search loop'
print *, '        i = ', i
print *, '        array(i) = ', array(i)
print *, '        exit loop D1 (ie, array(i) >= x)?', (array(i) >= x)
             if (array(i) >= x) exit
!!
                i = i + 1
             end do

             do                          ! Search from upper end
print *, ''
print *, '        start loop D2 - high search loop'
print *, '        j = ', j
print *, '        array(j) = ', array(j)
print *, '        exit loop D2 (ie, x >= array(j))?', (x >= array(j))
             if( x >= array(j)) exit
!!
                j = j - 1
             end do

print *, '      exit loop C (ie, i=',i,' > j=',j,')?', (i > j)
          if (i > j) exit

             ! Otherwise, swap positions i & j
print *, '      Otherwise, swap positions i & j'

             temp    = array(i)
             ww   = order(i)
             array(i) = array(j)
             order(i) = order(j)
             array(j) = temp
!!
             order(j) = ww
print *, '      increment i & j'
             i = i + 1
             j = j - 1
print *, '      exit loop C (ie, i=',i,' > j=',j,')?', (i > j)
          if (i > j) exit
          end do

          if (j-l >= r-i) then
print *, '      j-l=',j-l,' >= r-i=',r-i
             if (l < j) then
print *, '      l=',l,' < j=',j,': increment s'
                s = s + 1
                stackl(s) = l
                stackr(s) = j
print *, '      s now =', s
print *, '      stackl(s=',s,') = i =', i
print *, '      stackr(s=',s,') = r =', r
             end if
             l = i
          else
print *, '      j-l=',j-l,' < r-i=',r-i
             if (i < r) then
print *, '      i=',i,' < r=',r,': increment s'
                s = s + 1
                stackl(s) = i
                stackr(s) = r
print *, '      s now =', s
print *, '      stackl(s=',s,') = i =', i
print *, '      stackr(s=',s,') = r =', r
             end if
             r = j
          end if

print *, '    exit loop B (ie, l=',l,' >= r=',r,')?', (l >= r)
       if(l >= r) exit
       end do
print *, '  exit loop A (ie, s == 0)?', (s == 0)
    if (s == 0) exit
    end do

  end subroutine quickSortDoubleWithIndex_new

  !---------------------------------------------------------------------
  subroutine quickSortDoubleWithIndex(array, order)
  !---------------------------------------------------------------------

    real(double), intent(INOUT)    :: array(:)
    integer, intent(OUT) :: order(:)  ! returns original position
    real(double) :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(15), stackr(15), ww

    !-------------------------------------------------------------------
    ! Start

    do j = lbound(array,1),ubound(array,1)
       order(j) = j
    end do
    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
             if (array(i) >= x) exit
                i = i + 1
             end do

             do                          ! Search from upper end
             if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                ww   = order(i)
                array(i) = array(j)
                order(i) = order(j)
                array(j) = temp
                order(j) = ww
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
          exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
!    return

  end subroutine quickSortDoubleWithIndex

  !---------------------------------------------------------------------
  subroutine quickSortInt32WithIndex(array, order)
  !---------------------------------------------------------------------

    integer(int32), intent(INOUT)    :: array(:)
    integer, intent(OUT) :: order(:)  ! returns original position
    integer(int32)  :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(15), stackr(15), ww

    !-------------------------------------------------------------------
    ! Start

    do j = lbound(array,1),ubound(array,1)
       order(j) = j
    end do
    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                ww   = order(i)
                array(i) = array(j)
                order(i) = order(j)
                array(j) = temp
                order(j) = ww
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
!    return

  end subroutine quickSortInt32WithIndex

  !---------------------------------------------------------------------
  subroutine quickSortSingleWithIndex(array, order)
  !---------------------------------------------------------------------

    real(single), intent(INOUT)    :: array(:)
    integer, intent(OUT) :: order(:)  ! returns original position
    real(single)     :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(15), stackr(15), ww

    !-------------------------------------------------------------------
    ! Start

    do j = lbound(array,1),ubound(array,1)
       order(j) = j
    end do
    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                ww   = order(i)
                array(i) = array(j)
                order(i) = order(j)
                array(j) = temp
                order(j) = ww
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
!    return

  end subroutine quickSortSingleWithIndex

end module sort_mod
