!
! Module: ssc_misc
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: A grab-bag of general-purpose routines.
! 
! Comments: TODO:
! 
! $Id: ssc_misc.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module ssc_misc

  use types
  use constants

  implicit none 

  private 

  public nintAndTruncToInt8
  public nintAndTruncToInt16
  public nintAndTruncToInt32
  public fracVal
  public printEvery
  public sprintEvery
  public ucase
  public lcase
  public strToInt
  public intToStr
  public raDegToStr
  public decDegToStr
  public swap
  public fracDiff
  public getFreeIoUnit
  public removeTrailingChar0s
  public isOdd
  public int125
  public approxEqual
  public orLists
  public isNaN
  public calcMostSignifSetBit
  public calcLeastSignifSetBit

  interface nintAndTruncToInt8
    module procedure nintAndTruncToInt8Array, nintAndTruncToInt8Vector
  end interface

  interface nintAndTruncToInt16
    module procedure nintAndTruncToInt16Array, nintAndTruncToInt16Vector
  end interface

  interface nintAndTruncToInt32
    module procedure nintAndTruncToInt32Array, nintAndTruncToInt32Vector
  end interface

  interface fracVal
    module procedure fracValSingleScal2Scal,  fracValSingleVec2Scal
    module procedure fracValSingleScal2Vec,   fracValSingleVec2Vec
    module procedure fracValSingleScal2Array, fracValSingleArray2Scal
    module procedure fracValDoubleScal2Scal,  fracValDoubleVec2Scal
    module procedure fracValDoubleScal2Vec,   fracValDoubleVec2Vec
    module procedure fracValDoubleScal2Array, fracValDoubleArray2Scal

    module procedure fracValAugSingleScalar, fracValAugSingleVector
  end interface

  interface raDegToStr
    module procedure raDegToStrScalar
    module procedure raDegToStrVector
  end interface

  interface decDegToStr
    module procedure decDegToStrScalar
    module procedure decDegToStrVector
  end interface

  interface fracDiff
    module procedure fracDiffSingle
    module procedure fracDiffDouble
  end interface

  interface approxEqual
    module procedure approxEqualSingle
    module procedure approxEqualDouble
  end interface

!  logical :: firstCallPrintEvery = .true.
!  integer :: myJump ! used by printEvery

!**** why is it necessary to have these global? Can put them in the routine and give them the 'save' attribute??

contains 

  !---------------------------------------------------------------------
  function nintAndTruncToInt8Array(real32array) result(int8array)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: real32array(:,:)
    integer(int8) :: int8array(size(real32array, 1), size(real32array, 2))

    integer(int8) :: min, max

    !-------------------------------------------------------------------
    ! Start

    max = huge(1_int8)
    min = 1 - max
    where(real32array .ge. max)
      int8array = max
    elsewhere(real32array .le. min)
      int8array = min
    elsewhere
      int8array = nint(real32array, int8)
    end where

  end function nintAndTruncToInt8Array

  !---------------------------------------------------------------------
  function nintAndTruncToInt16Array(real32array) result(int16array)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: real32array(:,:)
    integer(int16) :: int16array(size(real32array, 1), size(real32array, 2))

    integer(int16) :: min, max

    !-------------------------------------------------------------------
    ! Start

    max = huge(1_int16)
    min = 1 - max
    where(real32array .ge. max)
      int16array = max
    elsewhere(real32array .le. min)
      int16array = min
    elsewhere
      int16array = nint(real32array, int16)
    end where

  end function nintAndTruncToInt16Array

  !---------------------------------------------------------------------
  function nintAndTruncToInt32Array(real32array) result(int32array)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: real32array(:,:)
    integer(int32) :: int32array(size(real32array, 1), size(real32array, 2))

    integer(int32) :: min, max

    !-------------------------------------------------------------------
    ! Start

    max = huge(1_int32)
    min = 1 - max
    where(real32array .ge. max)
      int32array = max
    elsewhere(real32array .le. min)
      int32array = min
    elsewhere
      int32array = nint(real32array, int32)
    end where

  end function nintAndTruncToInt32Array

  !---------------------------------------------------------------------
  function nintAndTruncToInt8Vector(real32Vector) result(int8Vector)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: real32Vector(:)
    integer(int8) :: int8Vector(size(real32Vector))

    integer(int8) :: min, max

    !-------------------------------------------------------------------
    ! Start

    max = huge(1_int8)
    min = 1 - max
    where(real32Vector .ge. max)
      int8Vector = max
    elsewhere(real32Vector .le. min)
      int8Vector = min
    elsewhere
      int8Vector = nint(real32Vector, int8)
    end where

  end function nintAndTruncToInt8Vector

  !---------------------------------------------------------------------
  function nintAndTruncToInt16Vector(real32Vector) result(int16Vector)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: real32Vector(:)
    integer(int16) :: int16Vector(size(real32Vector))

    integer(int16) :: min, max

    !-------------------------------------------------------------------
    ! Start

    max = huge(1_int16)
    min = 1 - max
    where(real32Vector .ge. max)
      int16Vector = max
    elsewhere(real32Vector .le. min)
      int16Vector = min
    elsewhere
      int16Vector = nint(real32Vector, int16)
    end where

  end function nintAndTruncToInt16Vector

  !---------------------------------------------------------------------
  function nintAndTruncToInt32Vector(real32Vector) result(int32Vector)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: real32Vector(:)
    integer(int32) :: int32Vector(size(real32Vector))

    integer(int32) :: min, max

    !-------------------------------------------------------------------
    ! Start

    max = huge(1_int32)
    min = 1 - max
    where(real32Vector .ge. max)
      int32Vector = max
    elsewhere(real32Vector .le. min)
      int32Vector = min
    elsewhere
      int32Vector = nint(real32Vector, int32)
    end where

  end function nintAndTruncToInt32Vector


  !---------------------------------------------------------------------
  real function fracValSingleScal2Scal(min, max, frac)
  !---------------------------------------------------------------------
  ! the double subtraction from 1 is to ensure that the multipliers
  ! of min and max add exactly to 1. If this were not done, numerical
  ! rounding effects can cause the result to be, for example, less than
  ! min for small positive values of frac.

    implicit none 

    real(single), intent(in) :: min, max, frac

    !-------------------------------------------------------------------
    ! Start:

    fracValSingleScal2Scal = (1.0 - frac) * min + (1.0 - (1.0 - frac)) * max

  end function fracValSingleScal2Scal

  !---------------------------------------------------------------------
  function fracValSingleVec2Scal(min, max, frac) result(fracValues)
  !---------------------------------------------------------------------
  ! the double subtraction from 1 is to ensure that the multipliers
  ! of min and max add exactly to 1. If this were not done, numerical
  ! rounding effects can cause the result to be, for example, less than
  ! min for small positive values of frac.

    implicit none 

    real(single), intent(in) :: min, max, frac(:)
    real(single) :: fracValues(size(frac))

    !-------------------------------------------------------------------
    ! Start:

    fracValues = (1.0 - frac) * min + (1.0 - (1.0 - frac)) * max

  end function fracValSingleVec2Scal

  !---------------------------------------------------------------------
  function fracValSingleScal2Vec(min, max, frac) result(fracValues)
  !---------------------------------------------------------------------
  ! the double subtraction from 1 is to ensure that the multipliers
  ! of min and max add exactly to 1. If this were not done, numerical
  ! rounding effects can cause the result to be, for example, less than
  ! min for small positive values of frac.

    implicit none 

    real(single), intent(in) :: min(:), max(size(min)), frac
    real(single) :: fracValues(size(min))

    !-------------------------------------------------------------------
    ! Start:

    fracValues = (1.0 - frac) * min + (1.0 - (1.0 - frac)) * max

  end function fracValSingleScal2Vec

  !---------------------------------------------------------------------
  function fracValSingleVec2Vec(min, max, frac) result(fracValues)
  !---------------------------------------------------------------------
  ! the double subtraction from 1 is to ensure that the multipliers
  ! of min and max add exactly to 1. If this were not done, numerical
  ! rounding effects can cause the result to be, for example, less than
  ! min for small positive values of frac.

    implicit none 

    real(single), intent(in) :: min(:), max(size(min)), frac(size(min))
    real(single) :: fracValues(size(min))

    !-------------------------------------------------------------------
    ! Start:

    fracValues = (1.0 - frac) * min + (1.0 - (1.0 - frac)) * max

  end function fracValSingleVec2Vec

  !---------------------------------------------------------------------
  function fracValSingleScal2Array(min, max, frac) result(fracValues)
  !---------------------------------------------------------------------
  ! the double subtraction from 1 is to ensure that the multipliers
  ! of min and max add exactly to 1. If this were not done, numerical
  ! rounding effects can cause the result to be, for example, less than
  ! min for small positive values of frac.

    implicit none 

    real(single), intent(in) :: min(:,:), max(size(min,1),size(min,2)), frac
    real(single) :: fracValues(size(min,1),size(min,2))

    !-------------------------------------------------------------------
    ! Start:

    fracValues = (1.0 - frac) * min + (1.0 - (1.0 - frac)) * max

  end function fracValSingleScal2Array

  !---------------------------------------------------------------------
  function fracValSingleArray2Scal(min, max, frac) result(fracValues)
  !---------------------------------------------------------------------
  ! the double subtraction from 1 is to ensure that the multipliers
  ! of min and max add exactly to 1. If this were not done, numerical
  ! rounding effects can cause the result to be, for example, less than
  ! min for small positive values of frac.

    implicit none 

    real(single), intent(in) :: min, max, frac(:,:)
    real(single) :: fracValues(size(frac,1),size(frac,2))

    !-------------------------------------------------------------------
    ! Start:

    fracValues = (1.0 - frac) * min + (1.0 - (1.0 - frac)) * max

  end function fracValSingleArray2Scal

  !---------------------------------------------------------------------
  real(double) function fracValDoubleScal2Scal(min, max, frac)
  !---------------------------------------------------------------------
  ! the double subtraction from 1 is to ensure that the multipliers
  ! of min and max add exactly to 1. If this were not done, numerical
  ! rounding effects can cause the result to be, for example, less than
  ! min for small positive values of frac.

    implicit none 

    real(double), intent(in) :: min, max, frac

    !-------------------------------------------------------------------
    ! Start:

    fracValDoubleScal2Scal = (1.0d0 - frac) * min&
      + (1.0d0 - (1.0d0 - frac)) * max

  end function fracValDoubleScal2Scal

  !---------------------------------------------------------------------
  function fracValDoubleVec2Scal(min, max, frac) result(fracValues)
  !---------------------------------------------------------------------
  ! the double subtraction from 1 is to ensure that the multipliers
  ! of min and max add exactly to 1. If this were not done, numerical
  ! rounding effects can cause the result to be, for example, less than
  ! min for small positive values of frac.

    implicit none 

    real(double), intent(in) :: min, max, frac(:)
    real(double) :: fracValues(size(frac))

    !-------------------------------------------------------------------
    ! Start:

    fracValues = (1.0d0 - frac) * min + (1.0d0 - (1.0d0 - frac)) * max

  end function fracValDoubleVec2Scal

  !---------------------------------------------------------------------
  function fracValDoubleScal2Vec(min, max, frac) result(fracValues)
  !---------------------------------------------------------------------
  ! the double subtraction from 1 is to ensure that the multipliers
  ! of min and max add exactly to 1. If this were not done, numerical
  ! rounding effects can cause the result to be, for example, less than
  ! min for small positive values of frac.

    implicit none 

    real(double), intent(in) :: min(:), max(size(min)), frac
    real(double) :: fracValues(size(min))

    !-------------------------------------------------------------------
    ! Start:

    fracValues = (1.0d0 - frac) * min + (1.0d0 - (1.0d0 - frac)) * max

  end function fracValDoubleScal2Vec

  !---------------------------------------------------------------------
  function fracValDoubleVec2Vec(min, max, frac) result(fracValues)
  !---------------------------------------------------------------------
  ! the double subtraction from 1 is to ensure that the multipliers
  ! of min and max add exactly to 1. If this were not done, numerical
  ! rounding effects can cause the result to be, for example, less than
  ! min for small positive values of frac.

    implicit none 

    real(double), intent(in) :: min(:), max(size(min)), frac(size(min))
    real(double) :: fracValues(size(min))

    !-------------------------------------------------------------------
    ! Start:

    fracValues = (1.0d0 - frac) * min + (1.0d0 - (1.0d0 - frac)) * max

  end function fracValDoubleVec2Vec

  !---------------------------------------------------------------------
  function fracValDoubleScal2Array(min, max, frac) result(fracValues)
  !---------------------------------------------------------------------
  ! the double subtraction from 1 is to ensure that the multipliers
  ! of min and max add exactly to 1. If this were not done, numerical
  ! rounding effects can cause the result to be, for example, less than
  ! min for small positive values of frac.

    implicit none 

    real(double), intent(in) :: min(:,:), max(size(min,1),size(min,2)), frac
    real(double) :: fracValues(size(min,1),size(min,2))

    !-------------------------------------------------------------------
    ! Start:

    fracValues = (1.0d0 - frac) * min + (1.0d0 - (1.0d0 - frac)) * max

  end function fracValDoubleScal2Array

  !---------------------------------------------------------------------
  function fracValDoubleArray2Scal(min, max, frac) result(fracValues)
  !---------------------------------------------------------------------
  ! the double subtraction from 1 is to ensure that the multipliers
  ! of min and max add exactly to 1. If this were not done, numerical
  ! rounding effects can cause the result to be, for example, less than
  ! min for small positive values of frac.

    implicit none 

    real(double), intent(in) :: min, max, frac(:,:)
    real(double) :: fracValues(size(frac,1),size(frac,2))

    !-------------------------------------------------------------------
    ! Start:

    fracValues = (1.0d0 - frac) * min + (1.0d0 - (1.0d0 - frac)) * max

  end function fracValDoubleArray2Scal



  !---------------------------------------------------------------------
  real function fracValAugSingleScalar(miny, maxy, minx, maxx, x)
  !---------------------------------------------------------------------
  ! NOTE!! This has NO protection against minx = maxx. Such a case may
  ! cause a seg fault.
  !
  ! the double subtraction from 1 is to ensure that the multipliers
  ! of min and max add exactly to 1. If this were not done, numerical
  ! rounding effects can cause the result to be, for example, less than
  ! min for small positive values of frac.

    implicit none 

    real(single), intent(in) :: miny, maxy, minx, maxx, x

    real(single) :: frac

    !-------------------------------------------------------------------
    ! Start:

    frac = (x - minx) / (maxx - minx)
    fracValAugSingleScalar = (1.0 - frac) * miny + (1.0 - (1.0 - frac)) * maxy

  end function fracValAugSingleScalar

  !---------------------------------------------------------------------
  function fracValAugSingleVector(miny, maxy, minx, maxx, x)&
  result(fracValues)
  !---------------------------------------------------------------------
  ! NOTE!! This has NO protection against minx = maxx. Such a case may
  ! cause a seg fault.
  !
  ! the double subtraction from 1 is to ensure that the multipliers
  ! of min and max add exactly to 1. If this were not done, numerical
  ! rounding effects can cause the result to be, for example, less than
  ! min for small positive values of frac.

    implicit none 

    real(single), intent(in) :: miny, maxy, minx, maxx, x(:)
    real(single) :: fracValues(size(x))

    real(single), allocatable :: frac(:)

    !-------------------------------------------------------------------
    ! Start:

    allocate(frac(size(x)))
    frac = (x - minx) / (maxx - minx)
    fracValues = (1.0 - frac) * miny + (1.0 - (1.0 - frac)) * maxy

  end function fracValAugSingleVector


  !---------------------------------------------------------------------
  subroutine printEvery(i, total, jump, fraction, message)
  !---------------------------------------------------------------------
  ! This is designed for monitoring the progress of large loops. In this
  ! situation a print statement that is called at every loop pass can
  ! generate a large amount of output and slow execution down a lot. The
  ! present subroutine is designed to print a message only every so often.
  ! The arguments 'jump' and 'fraction' are optional. If 'jump' is
  ! specified, the message is printed every 'jump' cycles. If 'jump' is
  ! not supplied but 'fraction' is, the jump number is calculated from
  ! fraction * total. If neither is given, the message is printed at
  ! every cycle.

    implicit none

    integer,      intent(in)           :: i, total
    integer,      intent(in), optional :: jump
    real(single), intent(in), optional :: fraction
    character(*), intent(in), optional :: message

    integer :: percent, numMilestones
    logical :: firstCall = .true.
    integer, allocatable, save :: mileStones(:)

    !-------------------------------------------------------------------
    ! Start

!    if (firstCallPrintEvery .and. present(fraction)) then
!      firstCallPrintEvery = .false.
!      myJump = min(total, max(1, nint(fraction * total)))
!    end if

    if (present(jump)) then
      if (modulo(i, jump) .eq. 0) then
        if (present(message)) then
          print *, message, i, '/', total
        else
          print *, i, '/', total
        end if
      end if
    else if (present(fraction)) then
      if (firstCall) then
        numMilestones = floor(1.0/fraction)
        allocate(mileStones(numMilestones))
        mileStones = (/(nint(i*total*fraction), i=1,numMilestones)/)
        firstCall = .false.
      end if

!      if (modulo(i, myJump) .eq. 0) then
      if (any(i == mileStones)) then
        percent = nint(100 * ((i - 1) / real(total - 1)))
!print *, 'i, total, percent:', i, total, percent
        if (present(message)) then
          print *, message, percent, '%'
        else
          print *, percent, '%'
        end if
!      else if (i >= total) then
!        firstCallPrintEvery = .true.
      end if

      if (i >= total) then
        firstCall = .true.
        deallocate(mileStones)
      end if

    else
      if (present(message)) then
        print *, message, i, '/', total
      else
        print *, i, '/', total
      end if
    end if

  end subroutine printEvery

  !---------------------------------------------------------------------
  function sprintEvery(i, total, jump, fraction, message) result(str)
  !---------------------------------------------------------------------
  ! See printEvery.

    implicit none

    integer,      intent(in)           :: i, total
    integer,      intent(in), optional :: jump
    real(single), intent(in), optional :: fraction
    character(*), intent(in), optional :: message
    character(256)                     :: str

    integer :: percent, numMilestones
    logical :: firstCall = .true.
    integer, allocatable, save :: mileStones(:)
    character(10) :: iStr, totalStr, percentStr

    !-------------------------------------------------------------------
    ! Start

    iStr = intToStr(i)
    totalStr = intToStr(total)
    str = ''

    if (present(jump)) then
      if (modulo(i, jump) .eq. 0) then
        if (present(message)) then
!          print *, message, i, '/', total
          write(str, '(a)') trim(message)//' '//trim(iStr)//'/'//trim(totalStr)
        else
!          print *, i, '/', total
          write(str, '(a)') trim(iStr)//'/'//trim(totalStr)
        end if
      end if
    else if (present(fraction)) then
      if (firstCall) then
        numMilestones = floor(1.0/fraction)
        allocate(mileStones(numMilestones))
        mileStones = (/(nint(i*total*fraction), i=1,numMilestones)/)
        firstCall = .false.
      end if

      if (any(i == mileStones)) then
        percent = nint(100 * ((i - 1) / real(total - 1)))
        percentStr = intToStr(percent)
        if (present(message)) then
!          print *, message, percent, '%'
          write(str, '(a)') trim(message)//' '//trim(percentStr)//' %'
        else
!          print *, percent, '%'
          write(str, '(a)') trim(percentStr)//' %'
        end if
      end if

      if (i >= total) then
        firstCall = .true.
        deallocate(mileStones)
      end if

    else
      if (present(message)) then
!        print *, message, i, '/', total
        write(str, '(a)') trim(message)//' '//trim(iStr)//'/'//trim(totalStr)
      else
!        print *, i, '/', total
        write(str, '(a)') trim(iStr)//'/'//trim(totalStr)
      end if
    end if

  end function sprintEvery

  !---------------------------------------------------------------------
  function ucase(arg) result(string)
  !---------------------------------------------------------------------

    implicit none 

    character(*), intent(in) :: arg
    character(len=len(arg)) :: string
    integer :: i, asciiIndex, aAsciiIndex, zAsciiIndex, bigAasciiIndex

    !-------------------------------------------------------------------
    ! Start:

    aAsciiIndex = ichar('a')
    zAsciiIndex = ichar('z')
    bigAasciiIndex = ichar('A')
    string = arg
    do i = 1, len(arg)
      asciiIndex = ichar(string(i:i))
      if (asciiIndex .ge. aAsciiIndex .and. asciiIndex .le. zAsciiIndex) then
        string(i:i) = char(bigAasciiIndex + asciiIndex - aAsciiIndex)
      end if
    end do

  end function ucase

  !---------------------------------------------------------------------
  function lcase(arg) result(string)
  !---------------------------------------------------------------------

    implicit none 

    character(*), intent(in) :: arg
    character(len=len(arg)) :: string
    integer :: i, asciiIndex, aAsciiIndex, bigAasciiIndex, bigZasciiIndex

    !-------------------------------------------------------------------
    ! Start:

    aAsciiIndex = ichar('a')
    bigAasciiIndex = ichar('A')
    bigZasciiIndex = ichar('Z')
    string = arg
    do i = 1, len(arg)
      asciiIndex = ichar(string(i:i))
      if (asciiIndex .ge. bigAasciiIndex&
      .and. asciiIndex .le. bigZasciiIndex) then
        string(i:i) = char(aAsciiIndex + asciiIndex - bigAasciiIndex)
      end if
    end do

  end function lcase

  !---------------------------------------------------------------------
  subroutine strToInt(str, int, status)
  !---------------------------------------------------------------------

    implicit none 

    character(*), intent(in) :: str
    integer, intent(out) :: int, status

    integer :: i, length, nonZeroLength, mult, digit

    !-------------------------------------------------------------------
    ! Start:

    status = 0 ! default
    int    = 0 ! default

    length = len(str)
    if (length .le. 0) then
      status = 1
  return
    end if

    do i = 1, length
      if (iachar(str(i:i)) .lt. iachar('0')&
      .or. iachar(str(i:i)) .gt. iachar('9')) then
        status = 2
  return
      end if
    end do

    nonZeroLength = 0
    do i = 1, length
      if (iachar(str(i:i)) .ne. iachar('0')) then
        nonZeroLength = length + 1 - i
    exit
      end if
    end do

    if (nonZeroLength .eq. 0) then
      int = 0
  return
    end if

    if (nonZeroLength .gt. floor(log10(real(huge(int))))) then
    ! floor because eg huge(int16) is 32768, thus ceiling(log10(huge(int16)))
    ! is 5, allowing a str value of, for example, '32769', which would cause
    ! an overflow.
      status = 3
  return
    end if

    int = iachar(str(length:length)) - iachar('0')
    mult = 1
    do i = length - 1, 1 + length - nonZeroLength, -1
      mult = mult * 10
      digit = iachar(str(i:i)) - iachar('0')
      int = int + mult * digit
    end do

  end subroutine strToInt


!***** define the integer type(s)?


  !---------------------------------------------------------------------
  function intToStr(int, leftJustify) result(str)
  !---------------------------------------------------------------------

    implicit none

    integer,       intent(in)  :: int
    logical(bool), intent(in), optional  :: leftJustify

    character(len=(2+range(int))) :: str

    logical(bool) :: localLeftJustify

    !-------------------------------------------------------------------
    ! Start

    if (present(leftJustify)) then
      localLeftJustify = leftJustify
    else
      localLeftJustify = .true.
    end if

    str = '' ! default

    write(str, *) int
    if (localLeftJustify) then
      str = adjustl(str)
    else
      str = adjustr(str)
    end if

  end function intToStr


!***** define the integer type(s)?


  !---------------------------------------------------------------------
  subroutine raDegToStrScalar(raDeg,  raStr)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: raDeg
    character(11), intent(out) :: raStr

    real(double) :: raDegVec(1)
    character(11) :: raStrVec(1)

    !-------------------------------------------------------------------
    ! Start

    raDegVec(1) = raDeg
    call raDegToStrVector(raDegVec, raStrVec)
    raStr = raStrVec(1)

  end subroutine raDegToStrScalar

  !---------------------------------------------------------------------
  subroutine raDegToStrVector(raDeg,  raStr)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: raDeg(:)
    character(11), intent(out) :: raStr(size(raDeg))

    real(double) :: raHours, raMin, raSec
    integer :: raHoursInt, raMinInt, i
    character(5) :: raSecStr

    !-------------------------------------------------------------------
    ! Start

    do i = 1, size(raDeg)
      raHours = raDeg(i) / 15.0
      raHoursInt = floor(raHours)
      raMin = (raHours - raHoursInt) * 60.0
      raMinInt = floor(raMin)
      raSec = (raMin - raMinInt) * 60.0

      if (raSec .lt. 1.0) then
        write(raSecStr, '(a2, f3.2)') '00', raSec
      else if (raSec .lt. 10.0) then
        write(raSecStr, '(a1, f4.2)') '0', raSec
      else
        write(raSecStr, '(f5.2)') raSec
      end if

      write(raStr(i), '(i2.2, a1, i2.2, a1, a5)') raHoursInt, ':', raMinInt&
      , ':', raSecStr
    end do

  end subroutine raDegToStrVector

  !---------------------------------------------------------------------
  subroutine decDegToStrScalar(decDeg,  decStr)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: decDeg
    character(12), intent(out) :: decStr

    real(double) :: decDegVec(1)
    character(12) :: decStrVec(1)

    !-------------------------------------------------------------------
    ! Start

    decDegVec(1) = decDeg
    call decDegToStrVector(decDegVec, decStrVec)
    decStr = decStrVec(1)

  end subroutine decDegToStrScalar

  !---------------------------------------------------------------------
  subroutine decDegToStrVector(decDeg, decStr)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: decDeg(:)
    character(12), intent(out) :: decStr(size(decDeg))

    real(double) :: absDecDeg, decMin, decSec
    integer :: absDecDegInt, decMinInt, i
    character(1) :: decSignStr
    character(5) :: decSecStr

    !-------------------------------------------------------------------
    ! Start

    do i = 1, size(decDeg)
      if(decDeg(i) .lt. 0.d0)then
        decSignStr = '-'
      else
        decSignStr = '+'
      endif

      absDecDeg = abs(decDeg(i))
      absDecDegInt = floor(absDecDeg)
      decMin = (absDecDeg - absDecDegInt) * 60.0
      decMinInt = floor(decMin)
      decSec = (decMin - decMinInt) * 60.0

      if (decSec .lt. 1.0) then
        write(decSecStr, '(a2, f3.2)') '00', decSec
      else if (decSec .lt. 10.0) then
        write(decSecStr, '(a1, f4.2)') '0', decSec
      else
        write(decSecStr, '(f5.2)') decSec
      end if

      write(decStr(i), '(a1, i2.2, a1, i2.2, a1, a5)') decSignStr, absDecDegInt&
      , ':', decMinInt, ':', decSecStr
    end do

  end subroutine decDegToStrVector

  !---------------------------------------------------------------------
  subroutine swap(a, b)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(inout) :: a, b
    real(single) :: temp

    !-------------------------------------------------------------------
    ! Start

    temp = a
    a = b
    b = temp

  end subroutine swap

  !---------------------------------------------------------------------
  function fracDiffSingle(a, b) result(fracDiff)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: a, b
    real(single) :: fracDiff

    !-------------------------------------------------------------------
    ! Start

    if (a + b == 0.0) then
      fracDiff = 0.0
  return
    end if

    fracDiff = 2.0 * abs((a - b) / (a + b))

  end function fracDiffSingle

  !---------------------------------------------------------------------
  function fracDiffDouble(a, b) result(fracDiff)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: a, b
    real(double) :: fracDiff

    !-------------------------------------------------------------------
    ! Start

    if (a + b == 0.0) then
      fracDiff = 0.0
  return
    end if

    fracDiff = 2.0d0 * abs((a - b) / (a + b))

  end function fracDiffDouble

  !---------------------------------------------------------------------
  subroutine getFreeIoUnit(unitNum)
  !---------------------------------------------------------------------
!**** status return????

    implicit none

    integer, intent(out) :: unitNum   ! unit number free for use.
    logical :: exists, open

    !-------------------------------------------------------------------
    ! Start

    do unitNum = 99, 0, -1
       inquire(unit=unitNum, exist=exists, opened=open)
  if (exists .and. .not. open) return
    end do

  end subroutine getFreeIoUnit

  !---------------------------------------------------------------------
  function removeTrailingChar0s(str) result(outStr)
  !---------------------------------------------------------------------

    implicit none 

    character(*), intent(in) :: str

    character(len=len(str)) :: outStr, tempstr
    character(1) :: chr
    integer :: length, i, index, lastChar

    !-------------------------------------------------------------------
    ! Start:

    length = len(str)
    if (length .le. 0) then
      outStr = str
  return
    end if

    lastChar = length
    do i = length, 1, -1
      chr = str(i:i)
      index = ichar(chr)
      if (index .ne. 0 .and. index .ne. ichar(' ')) then
        lastChar = i
    exit
      end if
      tempstr(i:i) = ' '
    end do
    tempstr(1:lastChar) = str(1:lastChar)

    outStr = tempstr

  end function removeTrailingChar0s

  !---------------------------------------------------------------------
  logical(bool) function isOdd(i)
  !---------------------------------------------------------------------

    implicit none 

    integer, intent(in) :: i

    !-------------------------------------------------------------------
    ! Start:

    if (real(floor(i / 2.0)) .eq. i / 2.0) then
      isOdd = .false.
    else
      isOdd = .true.
    end if

  end function isOdd

  !---------------------------------------------------------------------
  integer function int125(arg)
  !---------------------------------------------------------------------
  ! This function finds the value of 2, 5 or 10 which is nearest (in a
  ! logarithmic sense) to the input argument.

    implicit none

    real(single), intent(in) :: arg

    integer :: sign, mantissa
    real(single) :: log10Arg, logRemainder, localArg

    !-------------------------------------------------------------------
    ! Start

    if (arg .eq. 0.0) then
      int125 = 0
  return
    end if

    if (arg .lt. 0.0) then
      sign = -1
      localArg = -arg
    else
      sign = 1
      localArg = arg
    end if

    log10Arg = log10(localArg)
    mantissa = floor(log10Arg)
!**** test mantissa to guard against integer overflows.

    logRemainder = log10Arg - mantissa
    if (logRemainder .lt. lnTwo / lnTen / 2.0) then ! less than 0.150515,
    ! ie half-way between log10(1) and log10(2).

      int125 = 10**mantissa

    elseif (logRemainder .lt. 0.5) then ! less than 0.500000,
    ! ie half-way between log10(2) and log10(5).

      int125 = 2 * 10.0**mantissa

    elseif (logRemainder .lt. 1.0 - lnTwo / lnTen / 2.0) then ! less than
    ! 0.849485, ie half-way between log10(5) and log10(10).

      int125 = 5 * 10.0**mantissa

    else
      int125 = 10**(mantissa + 1)
    end if

    int125 = int125 * sign

  end function int125

  !---------------------------------------------------------------------
  logical(bool) function approxEqualSingle(argA, argB, precisionDegradation)
  !---------------------------------------------------------------------
  ! Well, we know that comparing reals is generally a bad idea, because
  ! numerical imprecision may introduce small differences between two
  ! real numbers that are supposed to be equal. Nevertheless it is useful
  ! to be able to perform such comparisons. The present function compares
  ! the two real arguments and returns true if they are equal to within a
  ! certain precision. The precision is that calculated for the 'real'
  ! data type, multiplied by precisionDegradation. (If the latter is not
  ! given, a default value of 5.0 is assumed.)

    implicit none

    real(single), intent(in)           :: argA, argB
    real(single), intent(in), optional :: precisionDegradation

    real(single) :: minAbsArg, criterion

    !-------------------------------------------------------------------
    ! Start

    minAbsArg = min(abs(argA), abs(argB))
    criterion = minAbsArg * 10.0**(-precision(minAbsArg))

    if (present(precisionDegradation)) then
      criterion = criterion * precisionDegradation
    else
      criterion = criterion * 5.0 ! default value
    end if

    if (abs(argA - argB) < criterion) then
      approxEqualSingle = .true.
    else
      approxEqualSingle = .false.
    end if

  end function approxEqualSingle

  !---------------------------------------------------------------------
  logical(bool) function approxEqualDouble(argA, argB, precisionDegradation)
  !---------------------------------------------------------------------
  ! Well, we know that comparing reals is generally a bad idea, because
  ! numerical imprecision may introduce small differences between two
  ! real numbers that are supposed to be equal. Nevertheless it is useful
  ! to be able to perform such comparisons. The present function compares
  ! the two real arguments and returns true if they are equal to within a
  ! certain precision. The precision is that calculated for the 'real'
  ! data type, multiplied by precisionDegradation. (If the latter is not
  ! given, a default value of 5.0 is assumed.)

    implicit none

    real(double), intent(in)           :: argA, argB
    real(double), intent(in), optional :: precisionDegradation

    real(double) :: minAbsArg, criterion

    !-------------------------------------------------------------------
    ! Start

    minAbsArg = min(abs(argA), abs(argB))
    criterion = minAbsArg * 10.0d0**(-precision(minAbsArg))

    if (present(precisionDegradation)) then
      criterion = criterion * precisionDegradation
    else
      criterion = criterion * 5.0d0 ! default value
    end if

    if (abs(argA - argB) < criterion) then
      approxEqualDouble = .true.
    else
      approxEqualDouble = .false.
    end if

  end function approxEqualDouble

  !---------------------------------------------------------------------
  function orLists(inList1, inList2) result(oredList)
  !---------------------------------------------------------------------
  ! This takes two lists of integers and returns their union.
  !
  ! The returned pointer oredList should be deallocated after use.
  !
  ! The function should NOT be called as in the following example:
  !
  !	oredList => orLists(oredList, list2)
  !
  ! The memory that the pointer oredList pointed to before the call now
  ! has no pointer to it, since at the moment the call is executed
  ! oredList is reassigned to the same memory that the function points
  ! to, which was newly assigned during the call. The initial memory
  ! pointed to by oredList therefore cannot now be deallocated. Better
  ! would be:
  !
  !	tempOredList => orLists(oredList, list2)
  !	deallocate(oredList)
  !	oredList => tempOredList
  !
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in) :: inList1(:), inList2(:)
    integer, pointer :: oredList(:)

    integer :: tempOredList(size(inList1) + size(inList2)),&
      size1, size2, numOred, i

    !-------------------------------------------------------------------
    ! Start

    size1 = size(inList1)
    size2 = size(inList2)
    tempOredList(1:size1) = inList1
    numOred = size1
    do i = 1, size2
      if (.not. any(tempOredList(:numOred) == inList2(i))) then
        numOred = numOred + 1
        tempOredList(numOred) = inList2(i)
      end if
    end do

    allocate(oredList(numOred))
    oredList = tempOredList(:numOred)

  end function orLists

  !---------------------------------------------------------------------
  logical(bool) function isNaN(arg)
  !---------------------------------------------------------------------
  ! NaNs don't seem to return true for any comparison, hence the test
  ! below. NOTE however that I haven't checked the f95 standard to see if
  ! this is a mandatory feature or just happens to be true on solaris-8.

    implicit none

    real(single), intent(in) :: arg

    !-------------------------------------------------------------------
    ! Start

    if(.not.(arg > 0.0) .and. .not.(arg <= 0.0)) then
      isNaN = .true.
    else
      isNaN = .false.
    end if

  end function isNaN

  !---------------------------------------------------------------------
  function calcMostSignifSetBit(i) result(msb)
  !---------------------------------------------------------------------

    implicit none

    integer(int32), intent(in) :: i
    integer                    :: msb

    integer :: bitNum

    !-------------------------------------------------------------------
    ! Start

    msb = -1 ! default
    do bitNum = 31, 0, -1
      if (btest(i, bitNum)) then
        msb = bitNum
  return
      end if
    end do

  end function calcMostSignifSetBit

  !---------------------------------------------------------------------
  function calcLeastSignifSetBit(i) result(lsb)
  !---------------------------------------------------------------------

    implicit none

    integer(int32), intent(in) :: i
    integer                    :: lsb

    integer :: bitNum

    !-------------------------------------------------------------------
    ! Start

    lsb = 32 ! default
    do bitNum = 0, 31
      if (btest(i, bitNum)) then
        lsb = bitNum
  return
      end if
    end do

  end function calcLeastSignifSetBit

  !---------------------------------------------------------------------
  character(32) function bitsToStr(argI)
  !---------------------------------------------------------------------

    implicit none

    integer(int32), intent(in) :: argI

    character(1) :: chars(32)
    integer :: i

    !-------------------------------------------------------------------
    ! Start

    chars = '0' ! default
    where(btest(argI, (/(i, i=31,0,-1)/))) chars = '1'

    bitsToStr =&
        chars( 1)//chars( 2)//chars( 3)//chars( 4)&
      //chars( 5)//chars( 6)//chars( 7)//chars( 8)&
      //chars( 9)//chars(10)//chars(11)//chars(12)&
      //chars(13)//chars(14)//chars(15)//chars(16)&
      //chars(17)//chars(18)//chars(19)//chars(20)&
      //chars(21)//chars(22)//chars(23)//chars(24)&
      //chars(25)//chars(26)//chars(27)//chars(28)&
      //chars(29)//chars(30)//chars(31)//chars(32)

  end function bitsToStr

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module ssc_misc
